(*
    This file is part of the Matrix project -
    which currently provides 4x4 matrices.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)


functor FMatrix4x4() :> MATRIX_4X4 =
  struct
    type tuples = Matrix4x4Type.t

    type t = tuples
    type scalar = real

    fun fromTuples m = m
    fun toTuples m = m

    (* When zeroValue is added to t, the result is t *)
    val zeroValue =
      ((0.0, 0.0, 0.0, 0.0),
       (0.0, 0.0, 0.0, 0.0),
       (0.0, 0.0, 0.0, 0.0),
       (0.0, 0.0, 0.0, 0.0))

    (* FIXME: Not tested! *)
    fun add ((amr0, amr1, amr2, amr3) : t,
             (bmr0, bmr1, bmr2, bmr3) : t) =
      let
        fun addRows((ac0, ac1, ac2, ac3), (bc0, bc1, bc2, bc3)) =
          (ac0+bc0, ac1+bc1, ac2+bc2, ac3+bc3)
      in
        (addRows(amr0, bmr0),
         addRows(amr1, bmr1),
         addRows(amr2, bmr2),
         addRows(amr3, bmr3))
      end

    (* Inverse number for add FIXME: Not tested! *)
    fun negate ((mr0, mr1, mr2, mr3) : t) = 
      let
        fun negateRow(c0, c1, c2, c3) =
          (~c0, ~c1, ~c2, ~c3)
      in
        (negateRow(mr0),
         negateRow(mr1),
         negateRow(mr2),
         negateRow(mr3))
      end

    (* FIXME: Not tested! *)  
    fun subtract (a, b) =
          add (a, negate b)

    (* FIXME: Not tested! *)
    fun equal ((amr0, amr1, amr2, amr3) : t,
               (bmr0, bmr1, bmr2, bmr3) : t) =
      let
        fun compareRows((ac0, ac1, ac2, ac3), (bc0, bc1, bc2, bc3)) =
          (Real.== (ac0, bc0) andalso
           Real.== (ac1, bc1) andalso
           Real.== (ac2, bc2) andalso
           Real.== (ac3, bc3))
      in
        (compareRows(amr0, bmr0) andalso
         compareRows(amr1, bmr1) andalso
         compareRows(amr2, bmr2) andalso
         compareRows(amr3, bmr3))
      end

    (* FIXME: Not tested! *)
    fun scale ((mr0, mr1, mr2, mr3) : t, factor) =
      let
        fun multRow(c0, c1, c2, c3) =
          (c0 * factor, c1 * factor, c2 * factor, c3 * factor)
      in
        (multRow(mr0),
         multRow(mr1),
         multRow(mr2),
         multRow(mr3))
      end

    (* FIXME: Not tested! *)
    fun inverseScale ((mr0, mr1, mr2, mr3) : t, divisor) =
      let
        fun divRow(c0, c1, c2, c3) =
          (c0 / divisor, c1 / divisor, c2 / divisor, c3 / divisor)
      in
        if Real.== (divisor, 0.0) then
          Excn.eRaiseGeneric ("ARITH", "Division by 0")
        else
          (divRow(mr0),
           divRow(mr1),
           divRow(mr2),
           divRow(mr3))
      end

    (* When unitValue is multiplied by t, the result is t *)
    val unitValue =
      ((1.0, 0.0, 0.0, 0.0),
       (0.0, 1.0, 0.0, 0.0),
       (0.0, 0.0, 1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))

    (* Used by multiply and other functions *)
    fun getIndex ((in0, in1, in2, in3), 0) = in0
      | getIndex ((in0, in1, in2, in3), 1) = in1
      | getIndex ((in0, in1, in2, in3), 2) = in2
      | getIndex ((in0, in1, in2, in3), 3) = in3
      | getIndex (_,_) =
          Excn.eRaiseBug ("fatal error in getIndex! Index outside 0-4!")

    fun multiply (matr1 : t, matr2 : t) =
      let
        val (mr0, mr1, mr2, mr3) = matr1;
        val (mrr0, mrr1, mrr2, mrr3) = matr2;
          
        fun multRowCol ((in0, in1, in2, in3), index) =
          ((in0 * getIndex(mrr0, index)) +
           (in1 * getIndex(mrr1, index)) +
           (in2 * getIndex(mrr2, index)) +
           (in3 * getIndex(mrr3, index)))

      in
        ((multRowCol(mr0, 0), multRowCol(mr0, 1), multRowCol(mr0, 2), multRowCol(mr0, 3)),
         (multRowCol(mr1, 0), multRowCol(mr1, 1), multRowCol(mr1, 2), multRowCol(mr1, 3)),
         (multRowCol(mr2, 0), multRowCol(mr2, 1), multRowCol(mr2, 2), multRowCol(mr2, 3)),
         (multRowCol(mr3, 0), multRowCol(mr3, 1), multRowCol(mr3, 2), multRowCol(mr3, 3)))
      end

    (* Used for calculating determinant
       This function would fit well in a Matrix3x3 module :) *)
    fun determinant3x3 ((a, b, c),
                        (d, e, f),
                        (g, h, i)) =
          a * (e * i - f * h) -
          b * (d * i - f * g) +
          c * (d * h - e * g) : real
    
    (* Determinant *)
    fun determinant ((a, b, c, d),
                     (e, f, g, h),
                     (i, j, k, l),
                     (m, n, O, p)) =
          a * (determinant3x3 ((f, g, h),
                               (j, k, l),
                               (n, O, p))) -
          b * (determinant3x3 ((e, g, h),
                               (i, k, l),
                               (m, O, p))) +
          c * (determinant3x3 ((e, f, h),
                               (i, j, l),
                               (m, n, p))) -
          d * (determinant3x3 ((e, f, g),
                               (i, j, k),
                               (m, n, O)))

    (* Used for calculating the adjoint *)
    fun cofactorRow (sign : real)
                    ((a, b, c, d),
                     (e, f, g, h),
                     (i, j, k, l)) =
          ( sign * (determinant3x3 ((b, c, d),
                                    (f, g, h),
                                    (j, k, l))),
           ~sign * (determinant3x3 ((a, c, d),
                                    (e, g, h),
                                    (i, k, l))),
            sign * (determinant3x3 ((a, b, d),
                                    (e, f, h),
                                    (i, j, l))),
           ~sign * (determinant3x3 ((a, b, c),
                                    (e, f, g),
                                    (i, j, k))))

    (* The transposed matrix *)
    fun transpose ((a, b, c, d),
                   (e, f, g, h),
                   (i, j, k, l),
                   (m, n, O, p)) =
          ((a, e, i, m),
           (b, f, j, n),
           (c, g, k, O),
           (d, h, l, p)) : t

    (* The adjoint of the matrix *)
    fun adjoint (mr0, mr1, mr2, mr3) =
          transpose
            (cofactorRow  1.0 (mr1, mr2, mr3),
             cofactorRow ~1.0 (mr0, mr2, mr3),
             cofactorRow  1.0 (mr0, mr1, mr3),
             cofactorRow ~1.0 (mr0, mr1, mr2))

    (* Inverse number for multiply *)
    fun invert (a : t) =
          let
            val det = determinant a
          in
            if Real.==(det, 0.0) then
              Excn.eRaiseGeneric
                ("DivideByZero", "Matrix4x4Mult.invert")
            else
              scale (adjoint a, 1.0 / det)
          end

    (* Divide is defined as multiply by the inverse *)
    fun divide (a : t, b : t) =
          multiply (a, invert b)

    (* All constant matrices such as rotation, scaling etc. *)
    structure Const = FMatrix4x4Const() :>
      MATRIX_4X4_COMMON where type t = t and
                              type angle = real

    (* Using matrices to transform geometry *)
    structure Trans = FMatrix4x4Transformation() :>
      MATRIX_4X4_TRANSFORMATION where type t = t
    
    (*structure Oper = ce3DMatrixOps (* All matrix operations such as rotate, scale etc. *)
    *)
  end

structure Matrix4x4 = FMatrix4x4()

structure Matrix4x4Test =
  struct
    (* Test determinant *)
    val m1 = ((2.0, 1.0, 3.0, 4.0),
              (0.0, 1.0, 1.0, 0.0),
              (1.0, 0.0, 1.0, 1.0),
              (4.0, 2.0, 1.0, ~1.0))
    val m1_4x4 = Matrix4x4.fromTuples m1

    fun testDeterminant () =
          if Real.== (Matrix4x4.determinant m1_4x4, ~10.0) then
            ()
          else
            print "There is a bug in Matrix4x4.determinant!"

    (* Test invert *)
    val m2 = ((1.0,0.0,2.0,0.0),
              (~1.0,3.0,0.0,0.0),
              (0.0,3.0,1.0,0.0),
              (0.0,0.0,0.0,1.0))
    val m2_4x4 = Matrix4x4.fromTuples m2;
    val im2_4x4 = Matrix4x4.invert m2_4x4;
    val im2 = Matrix4x4.toTuples im2_4x4;

    (* im2 should now be something like: 
       ((~1.0,~2.0,2.0,0.0),
        (~0.333333333333,~0.333333333333,0.666666666667,0.0),
        (1.0,1.0,~1.0,0.0),
        (0.0,0.0,0.0,1.0))
    *)
  end