(*
    This file is part of the Matrix project -
    which currently provides 4x4 matrices.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)


functor FMatrix4x4Const() =
  struct
    type t = Matrix4x4Type.t
    type angle = real

    fun rotateSinCosX (sinAngle : real, cosAngle : real) =
          ((1.0, 0.0, 0.0, 0.0),
           (0.0, cosAngle, ~sinAngle, 0.0),
           (0.0, sinAngle, cosAngle, 0.0),
           (0.0, 0.0, 0.0, 1.0))

    fun rotateX (radians : angle) =
          rotateSinCosX (Math.sin radians, Math.cos radians)
    
    fun rotateSinCosY (sinAngle : real, cosAngle : real) =
          ((cosAngle, 0.0, sinAngle, 0.0),
           (0.0, 1.0, 0.0, 0.0),
           (~sinAngle, 0.0, cosAngle, 0.0),
           (0.0, 0.0, 0.0, 1.0))

    fun rotateY (radians : angle) =
          rotateSinCosY (Math.sin radians, Math.cos radians)

    fun rotateSinCosZ (sinAngle : real, cosAngle : real) =
          ((cosAngle, ~sinAngle, 0.0, 0.0),
           (sinAngle, cosAngle, 0.0, 0.0),
           (0.0, 0.0, 1.0, 0.0),
           (0.0, 0.0, 0.0, 1.0))

    fun rotateZ (radians : angle) =
          rotateSinCosZ (Math.sin radians, Math.cos radians)

    (* FIXME: Not tested... normalize vector? *)
    fun rotateXYZ (radians, vx, vy, vz) =
          let
            val sinA = Math.sin radians
            val cosA = Math.cos radians
            val cosA1 = 1.0 - cosA
            val vx2 = vx * vx
            val vy2 = vy * vy
            val vz2 = vz * vz
          in
            ((vx2 + (1.0 - vx2) * cosA,
              vx * vy * cosA1 - vz * sinA,
              vz * vx * cosA1 + vy * sinA,
              0.0),
             (vx * vy * cosA1 + vz * sinA,
              vy2 + (1.0 - vy2) * cosA,
              vy * vz * cosA1 - vx * sinA,
              0.0),
             (vz * vx * cosA1 - vy * sinA,
              vy * vz * cosA1 + vx * sinA,
              vz2 + (1.0 - vz2) * cosA,
              0.0),
             (0.0, 0.0, 0.0, 1.0))
          end

    fun scaleXYZ (scaleX : real, scaleY : real, scaleZ : real) =
      ((scaleX, 0.0, 0.0, 0.0),
       (0.0, scaleY, 0.0, 0.0),
       (0.0, 0.0, scaleZ, 0.0),
       (0.0, 0.0, 0.0, 1.0))
      
    fun translateXYZ (translateX : real, translateY : real, translateZ : real) =
      ((1.0, 0.0, 0.0, translateX),
       (0.0, 1.0, 0.0, translateY),
       (0.0, 0.0, 1.0, translateZ),
       (0.0, 0.0, 0.0, 1.0))

    fun shearXY (shearX : real, shearY : real) =
      ((1.0, 0.0, shearX, 0.0),
       (0.0, 1.0, shearY, 0.0),
       (0.0, 0.0, 1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))
      
    fun shearXZ (shearX : real, shearZ : real) =
      ((1.0, shearX, 0.0, 0.0),
       (0.0, 1.0, 0.0, 0.0),
       (0.0, shearZ, 1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))
    
    fun shearYZ (shearY : real, shearZ : real) =
      ((1.0, 0.0, 0.0, 0.0),
       (shearY, 1.0, 0.0, 0.0),
       (shearZ, 0.0, 1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))

    (* FIXME: Not tested! *)
    fun perspective projPlaneDist =
          ((1.0, 0.0, 0.0, 0.0),
           (0.0, 1.0, 0.0, 0.0),
           (0.0, 0.0, 1.0, 0.0),
           (0.0, 0.0, 1.0 / projPlaneDist, 0.0))

    (* FIXME: Not tested! *)
    val parallel =
          ((1.0, 0.0, 0.0, 0.0),
           (0.0, 1.0, 0.0, 0.0),
           (0.0, 0.0, 0.0, 0.0),
           (0.0, 0.0, 0.0, 1.0))

    (* FIXME: Not tested! *)
    val swapLRCoordsys =
      ((1.0, 0.0, 0.0, 0.0),
       (0.0, 1.0, 0.0, 0.0),
       (0.0, 0.0, ~1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0)) : t
  end

(* FIXME: Make a complete testsuite!
structure ceTest3DMatrix : GN_TESTER =
  struct             
    structure matr = ce3DMatrix;
    
    fun translate () = 
          matr.equal(
            matr.translateXYZMatrix(1.6, 6.1, ~7.2),
            matr.translateXYZ(matr.unitValue ,1.6, 6.1, ~7.2))

    fun test () =
          gnTU.doTest [("translate", translate)]
  end *)