(*
    This file is part of the Matrix project -
    which currently provides 4x4 matrices.
    
    Copyright (C) 2000 ANOQ of the Sun (alias Johnny Andersen).
    
    E-mail: anoq@HardcoreProcessing.com

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    As a special exception, if you do not do anything which is not in
    the spirit of the GNU Library General Public License, you are not
    required to physically compile this software into a separate library,
    since this is generally not possible with current Stanard ML compilers.
    However if you do something which is not in the spirit of the
    GNU Library General Public License you will have to follow the
    licence perpetually - thus disallowing you to use it for any
    commercial purposes at all.

    If you are interested in a warranty or commercial support for this
    software, contact Hardcore Processing <sales@HardcoreProcessing.com>
    for more information.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*)


functor FMatrix4x4Transformation() =
  struct
    type t = Matrix4x4Type.t

    (* Return a function that transforms one 3D-point *)
    fun transformPoint (matr : t) {x, y, z} =
          let
            val ((in00, in01, in02, in03),
                 (in10, in11, in12, in13),
                 (in20, in21, in22, in23),
                 (in30, in31, in32, in33)) = matr
             
            val w = x * in30 + y * in31 + z * in32 + in33
          in
            if Real.== (0.0, w) then
              Excn.eRaiseBug ("Zero divide in Matrix4x4.Trans.transformPoint")
            else
              {x = ((in00 * x) + (in01 * y) + (in02 * z) + in03) / w,
               y = ((in10 * x) + (in11 * y) + (in12 * z) + in13) / w,
               z = ((in20 * x) + (in21 * y) + (in22 * z) + in23) / w}
          end

    (* Return a function that transforms one 3D-vector *)
    fun transformVector (matr : t) {x, y, z} =
          let
            val ((in00, in01, in02, _),
                 (in10, in11, in12, _),
                 (in20, in21, in22, _),
                 _) = matr
          in
            {x = (in00 * x) + (in01 * y) + (in02 * z),
             y = (in10 * x) + (in11 * y) + (in12 * z),
             z = (in20 * x) + (in21 * y) + (in22 * z)}
          end
  end
